package com.me.asteroids;

import com.badlogic.gdx.math.MathUtils;
import com.badlogic.gdx.math.Vector2;

import static com.me.asteroids.Constants.rand;

public final class AsteroidFactory {

    private int vertexCount;
    private float size;

    private float sizeVariation;
    private float angleVariation;

    private boolean sizeRelativeToLast;

    public AsteroidFactory setVertexCount(int vertexCount) {
        this.vertexCount = vertexCount;
        return this;
    }

    public AsteroidFactory setSize(float size) {
        this.size = size;
        return this;
    }

    public AsteroidFactory setSizeVariation(float sizeVariation) {
        this.sizeVariation = sizeVariation;
        return this;
    }

    public AsteroidFactory sizeRelativeToLast() {
        this.sizeRelativeToLast = true;
        return this;
    }

    public AsteroidFactory sizeRelativeToInitial() {
        this.sizeRelativeToLast = false;
        return this;
    }

    public AsteroidFactory setAngleVariation(float angleVariation) {
        this.angleVariation = angleVariation;
        return this;
    }

    private void validate() {
        if (vertexCount <= 2) {
            throw new IllegalStateException(String.format("Illegal vertexCount: %d. Must be >= 3.", vertexCount));
        }
        if (size <= 0) {
            throw new IllegalStateException(String.format("Illegal vertexCount: %f. Must be > 0.", size));
        }
        if (sizeVariation < 0) {
            throw new IllegalStateException(String.format("Illegal sizeVariation: %f. Must be >= 0.", sizeVariation));
        }
        if (sizeVariation > size) {
            throw new IllegalStateException(String.format("Illegal sizeVariation: %f. Must be <= size.", sizeVariation));
        }
        if (angleVariation < 0) {
            throw new IllegalStateException(String.format("Illegal angleVariation: %f. Must be >= 0.", angleVariation));
        }
        if (angleVariation > MathUtils.PI2 / vertexCount / 2) {
            throw new IllegalStateException(String.format("Illegal angleVariation: %f. May cause vertexes positions to swap.", angleVariation));
        }
    }

    private Vector2 applyAngleVariation(Vector2 vertex) {
        if (angleVariation > 0) {
            float half = angleVariation * 0.5f;
            vertex.rotateRad(rand.nextFloat(-half, half));
        }
        return vertex;
    }

    private float applySizeVariation(float size) {
        if (sizeVariation > 0) {
            float half = sizeVariation * 0.5f;
            float variation = rand.nextFloat(-half, half);
            if (sizeRelativeToLast) {
                size += variation;
                size = MathUtils.clamp(size, this.size - half, this.size + half);
            } else {
                size = this.size + variation;
            }
        }
        return size;
    }

    public float[] generate() {
        validate();

        float angleStep = MathUtils.PI2 / vertexCount;

        // Pick a random starting angle
        float startAngle = rand.nextFloat() * MathUtils.PI2;
        Vector2 dir = Utils.setUnitVectorAngleRad(new Vector2(), startAngle);

        float lastSize = size;
        float[] vertices = new float[vertexCount * 2];

        for (int i = 0; i < vertices.length; i += 2) {
            Vector2 vertex = dir.cpy();

            vertex = applyAngleVariation(vertex);
            lastSize = applySizeVariation(lastSize);

            vertex.scl(lastSize);
            vertices[i] = vertex.x;
            vertices[i + 1] = vertex.y;

            dir.rotateRad(angleStep);
        }

        return vertices;
    }

}
