package com.me.common.ecs;

import com.badlogic.gdx.utils.Array;
import com.badlogic.gdx.utils.ObjectMap;

final class EntityManager {

    private final Engine engine;

    private final Array<Entity> entities;
    private final Array<Entity> removedEntities;
    private final Array<Entity> toRefresh;

    private final ObjectMap<String, Array<Entity>> groupEntities;
    private final ObjectMap<String, Entity> taggedEntities;

    EntityManager(Engine engine) {
        this.engine = engine;

        this.entities = new Array<>(false, 16);
        this.removedEntities = new Array<>(false, 16);
        this.toRefresh = new Array<>(false, 16);

        this.groupEntities = new ObjectMap<>();
        this.taggedEntities = new ObjectMap<>();
    }

    Array<Entity> getEntities() {
        return entities;
    }

    Entity create() {
        Entity entity;
        if (!removedEntities.isEmpty()) {
            entity = removedEntities.removeIndex(0);
        } else {
            entity = new Entity(engine);
        }
        entity.reset();
        entities.add(entity);
        return entity;
    }

    void update() {
        if (toRefresh.isEmpty()) {
            return;
        }

        for (Entity entity : toRefresh) {
            refreshEntity(entity);
            entity.pendingRefresh = false;
        }

        toRefresh.clear();
    }

    Array<Entity> getGroupEntities(String group) {
        Array<Entity> entities = groupEntities.get(group);
        if (entities == null) {
            entities = new Array<>();
            groupEntities.put(group, entities);
        }
        return entities;
    }

    void addEntityToGroup(Entity entity, String group) {
        if (!entity.groups.contains(group)) {
            getGroupEntities(group).add(entity);
            entity.groups.add(group);
        }
    }

    void removeEntityFromGroup(Entity entity, String group) {
        getGroupEntities(group).removeValue(entity, true);
        entity.groups.remove(group);
    }

    Entity getEntityByTag(String tag) {
        return taggedEntities.get(tag);
    }

    void tagEntity(Entity entity, String tag) {
        Entity previouslyTagged = taggedEntities.put(tag, entity);
        if (previouslyTagged != null) {
            previouslyTagged.tag = null;
        }
        entity.tag = tag;
    }

    void removeEntityTag(Entity entity) {
        if (entity.tag != null) {
            taggedEntities.remove(entity.tag);
            entity.tag = null;
        }
    }

    void queueRefresh(Entity entity) {
        toRefresh.add(entity);
    }

    private void refreshEntity(Entity entity) {
        for (BaseSystem system : engine.getSystems()) {
            if (!(system instanceof EntitySystem)) {
                continue;
            }

            ((EntitySystem) system).refresh(entity);
        }

        if (entity.isRemoved()) {
            engine.removeAllEntityComponents(entity.id);
            entities.removeValue(entity, true);
            removedEntities.add(entity);

            // remove the entity from any groups it is in
            for (String group : entity.groups) {
                getGroupEntities(group).removeValue(entity, true);
            }
            entity.groups.clear();

            removeEntityTag(entity);
        }
    }


}
